
--- True color library for Euphoria
--- Alpha release
--- By Christopher Street.
--- Questions? Comments? Bugs? Send them to: chris_street@usa.net


without warning
without type_check
include machine.e

sequence RM_struc
global sequence SVGA_info,mode_info
atom struc_add,sim_RM_add,VESA_add,sprite_add
integer VESAinit,current_bank,granularity,raster,bpp,bank_segment,yres,
        num_banks

VESAinit=0              mode_info={}        bpp=4
struc_add=allocate(50)
sprite_add=allocate(71)

constant sim_RM_code={
                      #50,#53,#51,#57,          -- push eax,ebx,ecx,edi
                      #B8,#00,#03,#00,#00,      -- mov eax,0x300
                      #B3,#10,                  -- mov bl,0x10
                      #B7,#01,                  -- mov bh,1
                      #31,#C9,                  -- xor ecx,ecx
                      #BF}&int_to_bytes(struc_add)&  -- mov edi,offset buffer
                     {#CD,#31,                  -- int 0x31
                      #5F,#59,#5B,#58,          -- pop edi,ecx,ebx,eax
                      #C3                       -- ret
                     },
         sprite_code={
                      #51,#50,#53,#57,          -- push ecx,eax,ebx,edi
                      #8B,#3D}&int_to_bytes(sprite_add+#3B)&
                     {#8B,#1D}&int_to_bytes(sprite_add+#3F)&
                     {#8B,#0D}&int_to_bytes(sprite_add+#43)&
                     {#8B,#07,                  -- again: mov eax,[edi]
                      #C1,#E0,#08,              -- shl eax,8
                      #C1,#E8,#08,              -- shr eax,8
                      #3D,#00,#00,#00,#00,      -- cmp eax,0
                      #74,#02,                  -- jz skip
                      #89,#03,                  -- mov [ebx],eax
                      #81,#C3,#04,#00,#00,#00,  -- skip: add ebx,4
                      #81,#C7,#04,#00,#00,#00,  -- add edi,4
                      #49,                      -- dec ecx
                      #75,#E0,                  -- jnz again
                      #5F,#5B,#58,#59,          -- pop edi,ebx,eax,ecx
                      #C3,                      -- ret
                      #00,#00,#00,#00,
                      #00,#00,#00,#00,
                      #00,#00,#00,#00
                     }

global constant ERROR_NO_VESA=1,
                ERROR_MODE_NOT_SUPPORTED=1,
                TRUE_320X200=#10F,
                TRUE_640X480=#112,
                TRUE_800X600=#115,
                TRUE_1024X768=#118,
                TRUE_1280X1024=#11B,
                TARGA_UNKNOWN_FORMAT=1,
                TARGA_OUT_OF_MEMORY=2


global function init_VESA()
    sim_RM_add=allocate(length(sim_RM_code))
    poke(sim_RM_add,sim_RM_code)
    poke(sprite_add,sprite_code)
    VESA_add=allocate_low(512)

    RM_struc=int_to_bytes(remainder(VESA_add,65536))&{0,0,0,0,0,0,0,0,0,0,0,0,
         0,0,0,0,0,0,0,0,#12,#1,0,0,#00,#4F,0,0,0,0,0,floor(VESA_add/65536)*16,
         0,0,0,0,0,0,0,0,0,0,0,0,0,0}

    poke(struc_add,RM_struc)
    call(sim_RM_add)

    SVGA_info=peek({VESA_add,512})    
    if compare(SVGA_info[1..4],"VESA") then
        return ERROR_NO_VESA        --- VESA's not supported. What a pity!
    else
        VESAinit=1
        return 0                    --- All's well!
    end if
end function

global function convert_bitmap(sequence pic)
    integer i
    for e=1 to length(pic[1]) do
        i=pic[1][e][3]
        pic[1][e][3]=pic[1][e][1]
        pic[1][e][1]=i
    end for
    for e=1 to length(pic[2]) do
        for f=1 to length(pic[2][e]) do
            pic[2][e][f]=pic[1][pic[2][e][f]+1]
        end for
    end for
    return pic[2]
end function

global function bitmap2virtual(sequence pic)
    atom a,b
    a=allocate((length(pic[1])*length(pic)*bpp)+8)
    if not a then
        return a
    end if
    poke4(a,{length(pic[1]),length(pic)})
    b=a+8
    for e=1 to length(pic) do
        for f=1 to length(pic[e]) do
            if bpp=4 then
                pic[e][f]=pic[e][f]&0
            end if
            poke(b,pic[e][f])
            b=b+bpp
        end for
    end for
    return a
end function

function get_bytes(integer handle,integer number)
    sequence s
    integer i
    s={}
    for e=1 to number do
        i=getc(handle)
        if i=-1 then
            return s
        else
            s=s&i
        end if
    end for
    return s
end function

function tar16to24(sequence bytes)
    integer i
    sequence s
    s={}
    s=s&and_bits(bytes[1],124)*2
    s=s&(and_bits(bytes[1],3)*8)+(and_bits(bytes[2],224)/32)*8
    s=s&(and_bits(bytes[2],31)*8)
    return s
end function

global function read_targa(sequence filename)
    integer handle,ident_field,color_map,image_type,i,width,height,raster,j
    sequence map_specs,pic_specs,data,c_map
    atom a,b
    handle=open(filename,"rb")
    if handle=-1 then
        return 0
    end if
    ident_field=getc(handle)
    color_map=getc(handle)
    image_type=getc(handle)
    map_specs=get_bytes(handle,5)
    pic_specs=get_bytes(handle,10)
    if ident_field then
        data=get_bytes(handle,ident_field)
    end if
    if color_map then
        i=(map_specs[5]/16)*(map_specs[3]+map_specs[4]*256)
        if image_type=2 then
            data=get_bytes(handle,i)
        else
            c_map={}
            for e=1 to map_specs[3]+map_specs[4]*256 do
                c_map=append(c_map,get_bytes(handle,3))
                if i=4 then
                    data={getc(handle)}
                end if
            end for
        end if
    end if
    i=(pic_specs[9]/8)
    width=pic_specs[5]+pic_specs[6]*256
    height=pic_specs[7]+pic_specs[8]*256
    if image_type=1 then
        data=get_bytes(handle,width*height)
        a=allocate((bpp*width*height)+8)
        b=a
        if not a then
            return TARGA_OUT_OF_MEMORY
        end if
        poke4(a,{width,height})
        a=a+8
        if i=4 then
            for e=1 to length(data) do
                poke(a,c_map[data[e]+1])
                a=a+4
            end for        
        end if
    elsif image_type=2 then
        data=get_bytes(handle,i*width*height)
        a=allocate((bpp*width*height)+8)
        b=a
        if not a then
            return TARGA_OUT_OF_MEMORY
        end if
        poke4(a,{width,height})
        a=a+8
        if i=2 then
            for e=1 to length(data) by i do
                poke(a,tar16to24(data[e..e+1])&0)
                a=a+bpp
            end for
        elsif i=3 then
            for e=1 to length(data) by 3 do
                poke(a,data[e..e+2]&0)
                a=a+bpp
            end for
        elsif i=4 then
            for e=1 to length(data) by 4 do
                poke(a,data[e..e+3])
                a=a+bpp
            end for        
        end if
    else
        return TARGA_UNKNOWN_FORMAT
    end if
    close(handle)
    return b
end function

global function set_true_mode(integer VESAmode)
    if not VESAinit then
        return ERROR_NO_VESA
    end if
    RM_struc[29]=#01
    RM_struc[25..26]={remainder(VESAmode,256),floor(VESAmode/256)}
    poke(struc_add,RM_struc)
    call(sim_RM_add)
    mode_info=peek({VESA_add,256})
    granularity=(mode_info[5]+mode_info[6]*256)*1024
    if mode_info[26] then
        if mode_info[26]=24 then
            bpp=3
            poke(sprite_add+41,3)
            poke(sprite_add+47,3)
        else
            bpp=4
            poke(sprite_add+41,4)
            poke(sprite_add+47,4)
        end if
    else
        bpp=4
        poke(sprite_add+41,4)
        poke(sprite_add+47,4)
    end if
    bank_segment=(mode_info[9]+mode_info[10]*256)*16
    if VESAmode=#10F then
        raster=320      yres=200
    elsif VESAmode=#112 then
        raster=640      yres=480
    elsif VESAmode=#115 then
        raster=800      yres=600
    elsif VESAmode=#118 then
        raster=1024     yres=768
    elsif VESAmode=#11B then
        raster=1280     yres=1024
    else
        return ERROR_MODE_NOT_SUPPORTED
    end if
    if not and_bits(mode_info[1],1) then
        return ERROR_MODE_NOT_SUPPORTED
    end if    
    num_banks=floor((yres*raster*bpp)/granularity)
    RM_struc[29]=#02
    RM_struc[17..18]={remainder(VESAmode,256),floor(VESAmode/256)}
    poke(struc_add,RM_struc)
    call(sim_RM_add)
    current_bank=0
    return 0
end function

procedure set_bank(integer bank)
    RM_struc[21..22]={remainder(bank,256),floor(bank/256)}
    RM_struc[17..18]={0,0}
    RM_struc[29]=#05
    poke(struc_add,RM_struc)
    call(sim_RM_add)
    current_bank=bank
end procedure

global function get_dimensions(atom a)
    return peek4u({a,2})
end function

global function get_width(atom a)
    return peek4u(a)
end function

global function get_height(atom a)
    return peek4u(a+4)
end function

global procedure true_pixel(sequence where,sequence bgr)
    integer i,j
    i=((where[2]*raster)+where[1])*bpp
    j=current_bank*granularity
    if not sequence(bgr[1]) then
        bgr={bgr}
    end if
    if i>j+65535 or i<j then
        set_bank(floor(i/granularity))
        j=current_bank*granularity
    end if
    for e=1 to length(bgr) do
        if i>j and i<j+65535 then
            poke(bank_segment+i-j,bgr[e])
        else
            set_bank(floor(i/granularity))
            j=current_bank*granularity
            poke(bank_segment+i-j,bgr[e])
        end if
        i=i+bpp
    end for
end procedure

global procedure set_sprite(atom a1,atom a2,sequence where)
    integer w1,h1,nx,ny,w2,h2
    atom sp,rf,s
    w1=get_width(a1)    h1=get_height(a1)
    w2=get_width(a2)    h2=get_height(a2)
    if where[1]<w2 and where[2]<h2 then
        sp=a2+8+((where[2]*w2)+where[1])*bpp
        if w1+where[1]>w2 then
            nx=w2-where[1]
        else
            nx=w1
        end if
        if h1+where[2]>h2 then
            ny=h2-where[2]
        else
            ny=h1
        end if
        rf=a1+8
        poke4(sprite_add+67,nx)
        for e=1 to ny do
            poke4(sprite_add+59,{rf,sp})
            call(sprite_add)
            rf=rf+w1*bpp
            sp=sp+w2*bpp
        end for
    end if
end procedure

global function read_pixel(sequence where,integer n)
    sequence s
    integer i,j
    s={}
    j=current_bank*granularity
    i=((where[2]*raster)+where[1])*bpp
    if i<j or i>j+65535 then
        set_bank(floor(i/granularity))
        j=current_bank*granularity
    end if
    for e=1 to n do
        s=append(s,peek({bank_segment+i-j,3}))
        i=i+bpp
        if i>j+65535 then
            set_bank(floor(i/granularity))
            j=current_bank*granularity
        end if
    end for
    return s
end function

global function read_virtual_pixel(atom a,sequence where,integer n)
    sequence s
    s={}
    a=a+8+((where[2]*get_width(a))+where[1])*bpp
    for e=1 to n do
        s=append(s,peek({a,3}))
        a=a+bpp
    end for
    return s
end function

global procedure virtual_pixel(atom a,sequence where,sequence bgr)
    integer x,y,nx
    x=peek4u(a)
    x=x*bpp
    y=peek4u(a+4)
    if not sequence(bgr[1])=1 then
        bgr={bgr}
    end if
    if where[1]<x and where[2]<y then
        a=a+8+(where[2]*x)+where[1]*bpp
        x=x/4
        if where[1]+length(bgr)>x then
            nx=x-where[1]
        else
            nx=length(bgr)
        end if
        for e=1 to nx do
            poke(a,bgr[e])
            a=a+bpp
        end for
    end if
end procedure

global procedure display_true_image(sequence where,sequence image)
    for e=1 to length(image) do
        true_pixel(where+{0,e-1},image[e])
    end for
end procedure

global procedure copy_image(atom a1,atom a2,sequence w1,sequence w2)
    atom b1,b2
    integer nx,ny,wid1,wid2
    b1=a1+8+((w1[1][2]*get_width(a1))+w1[1][1])*bpp
    b2=a2+8+((w2[2]*get_width(a2))+w2[1])*bpp
    nx=w1[2][1]-w1[1][2]+1
    ny=w1[2][2]-w1[1][2]+1
    wid1=get_width(a1)      wid2=get_width(a2)
    if nx+w2[1]>=wid2 then
        nx=wid2-w2[1]
    end if
    if ny+w2[2]>=get_height(a2) then
        ny=get_height(a2)-w2[2]
    end if
    for e=1 to ny do
        mem_copy(b2,b1,nx*bpp)
        b1=b1+(wid1*bpp)
        b2=b2+(wid2*bpp)
    end for
end procedure

global function make_virtual_screen(integer x,integer y,sequence c)
    atom a
    a=allocate((x*y*bpp)+8)
    if bpp=4 then   c=c&0 end if
    if not a then
        return a
    end if
    poke4(a,{x,y})
    if length(c)=bpp then
        a=a+8
        for e=1 to x do
            for f=1 to y do
                poke(a,c)
                a=a+4
            end for
        end for
        return a-(x*y*bpp)-8
    end if
    return a
end function

global function virtual2bitmap(atom a1)
    sequence pic
    integer x,y
    x=get_width(a1)     y=get_height(a1)
    pic=repeat(repeat({},x),y)
    a1=a1+8
    for e=1 to y do
        for f=1 to x do
            pic[e][f]=peek({a1,3})
            a1=a1+bpp
        end for
    end for
    return pic
end function

global procedure clear_true_screen()
    for e=0 to num_banks do
        set_bank(e)
        mem_set(bank_segment,0,65536)
    end for
end procedure

global procedure set_virtual_screen(sequence where,sequence w2,atom a)
    integer x,y,j,n
    atom b
    if not length(w2) then
        w2={{0,0},{get_width(a)-1,get_height(a)-1}}
    end if
    x=peek4u(a)     n=w2[2][1]-w2[1][1]+1
    x=x*bpp
    y=w2[2][2]-w2[1][2]+1
    a=a+8
    a=a+(w2[1][2]*x)+(w2[1][1]*bpp)
    b=((where[2]*raster)+where[1])*bpp
    j=current_bank*granularity
    if where[1]+n>raster then
        n=raster-where[1]
    end if
    n=n*bpp
    if where[2]+y>yres then
        y=yres-where[2]
    end if
    for e=1 to y do
        if b<j or b>(j+65535) then
            set_bank(floor(b/granularity))
            j=current_bank*granularity
        end if
        if b+n>j+65536 then
            mem_copy(bank_segment+b-j,a,-b+j+65536)
            set_bank(current_bank+1)
            j=current_bank*granularity
            mem_copy(bank_segment,a-b+j,n+b-j)
        else
            mem_copy(bank_segment+b-j,a,n)
        end if
        a=a+x
        b=b+(raster*bpp)
    end for
end procedure



